/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * dheap.c
 *
 * Semi-dynamic memory management
 *
 * Author:  Michael Goldwasser
 * Modified:  Jan 1, 1995 --- Original release (Goldwasser)
 *******************************************************/

#include <stdio.h>
#include "dheap.h"
#include "standards.h"


/***********************************************************
              Dynamic Heap (dheap)
 ***********************************************************/

typedef struct Dblock {
    void		*base;
    unsigned int	open;	/* offset from base...keep aligned! */
    struct Dblock	*next;
} Dblock;

typedef struct Dheap {
    unsigned int	size;	/* size of last block created */
    Dblock		*block;	/* current is front of list */
} Dheap;



Dheap*  Dheap_New(init_size)
  unsigned int	init_size;	/* if 0 --> use default */
{
    Dheap	*d;

    d = ALLOCATE(Dheap);
    d->size = ((init_size > 0) ? init_size : DHEAP_INIT_SIZE);
    d->block = ALLOCATE(Dblock);
    d->block->next = NULL;
    d->block->base = (void*) MALLOC(d->size);
    d->block->open = 0;

    return(d);
}


void*	Dheap_Alloc(d,newsize)
  Dheap*	d;
  unsigned int	newsize;
{
    void*	ptr;
    Dblock*	newblock=NULL;

    if ((d->size - d->block->open) < newsize) {
	d->size = (unsigned int) ( ((float) d->size) * DHEAP_GEOMETRIC);
	while (d->size < newsize) {
	    d->size = (unsigned int) ( ((float) d->size) * DHEAP_GEOMETRIC);
	}
	newblock = ALLOCATE(Dblock);
	while ((newblock->base = (void*) MALLOC(d->size))==NULL) {
	    if (d->size <= newsize) {
		printf("Dheap_Alloc:  out of memory.\n");
		return(NULL);
	    } else
		printf("Dheap_Alloc:  memory running out.\n");
	    d->size =
		(unsigned int) ( ((float) d->size) / DHEAP_GEOMETRIC);
	    if (d->size < newsize)  d->size = newsize;
	}
	newblock->open = 0;
	newblock->next = d->block;
	d->block = newblock;
    }
    
    /**  Now, we're sure there's room **/
    ptr = (void *) (((unsigned long)d->block->base) + d->block->open);
    d->block->open += (newsize);
    if ((newsize % DHEAP_ALIGN)!=0)
	d->block->open += (DHEAP_ALIGN - (newsize % DHEAP_ALIGN));

    return(ptr);
}


Dheap_Freeall(d)
  Dheap*	d;
{
    Dblock	*trav1,*trav2;

    for (trav1=d->block; trav1!=NULL; trav1=trav2) {
	trav2=trav1->next;
	FREE(trav1->base);
	FREE(trav1);
    }
    FREE(d);
}
