/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * dheap.h
 *
 * Semi-dynamic memory management
 *
 * Author:  Michael Goldwasser
 * Modified:  Jan 1, 1995 --- Original release (Goldwasser)
 *******************************************************/

/*********************************************************************


   **********
      Dheap
   **********
   A dheap is a basic memory manager that sits on top of the system
   memory manager.  This takes advantage of groups of objects that do
   not need to be freed individually.  By grabbing large continuous
   blocks of memory at once for related variable, the locality of
   reference will be utilized, and the number of system calls will
   also be reduced.

   The Dheap simply grabs a chuck of memory and hands it out
   individually to callers.  When it runs out of memory, it grabs a
   new chunk twice as large as the previous one if possible, and it
   keeps a list of all the chunks so that they can all be freed later
   on.

**********************************************************************/

#ifndef _DHEAP
#define _DHEAP

#define DHEAP_ALIGN 4
#define DHEAP_INIT_SIZE 256
#define DHEAP_GEOMETRIC 2.0

struct Dheap* Dheap_New(unsigned int);
void*  Dheap_Alloc(struct Dheap*, unsigned int);
int    Dheap_Freeall(struct Dheap*);

#endif
