/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/***********************************************************************/
/**                                                                   **/
/**                            lists.h                                **/
/**                 Functions for operating on lists.                 **/
/**                                                                   **/
/**                   by Bruce Romney, October 1992                   **/
/**             (based on the ACT solid modelling package)            **/
/**                                                                   **/
/***********************************************************************/

#ifndef _LISTS
#define _LISTS

#include "standards.h"	/* contains the line defining 'List' as */
                        /* as a pointer to 'struct Element'     */

/*-------------------------------------------------------------*/
/*    Define generic pointer   		       		       */
/*-------------------------------------------------------------*/
typedef void *Ptr;

/*-------------------------------------------------------------*/
/*    Define list					       */
/*-------------------------------------------------------------*/
typedef struct Element
{
   List next;		/* next element in the list */
   List prev;		/* previous element in the list */
   Ptr item;		/* pointer to contents of this element */
} Element;

/*----------------------------------------------------------------*/
/*   Macros for using lists					  */
/*----------------------------------------------------------------*/
#define MAKE_ELT (List) (MALLOC(sizeof(struct Element)))
#define NXT(l) ((l)->next)
#define PRV(l) ((l)->prev)
#define GET0(l) ((l)->item)
#define GET(elt,type) ((type *)((elt)->item))
#define POP(l,type) ((type *)pop_element(&(l)))

/*----------------------------------------------------------------*/
/*                Function Declarations                           */
/*----------------------------------------------------------------*/
List init_list(Ptr);
List last(List);
List first(List);
Boolean memberp(Ptr, List);
Boolean firstp(List);
Boolean lastp(List);
int length(List);
List push(List, Ptr);
List add_to_end(List, Ptr);
List concatenate(List, List);
void set(List, Ptr);
List copy_list(List, List);
List reverse(List, List);
void delete_list(List);
void delete_list_of_lists(List);
List delete_elt(List, List);
List delete_obj(Ptr, List);
Ptr pop_element(List *);
List clip(List, List);

#endif
